#© 2020, The Sage Group plc or its licensors. All rights reserved
[CmdletBinding(SupportsShouldProcess = $true)]
param(
	[parameter()]
	[ValidateSet('64-bit', '32-bit')]
	[string]$Architecture = '64-bit',

	[switch]$EnableContextMenus,

	[switch]$SkipFolderInstall,

	[switch]$SkipNPMInstall,

	[switch]$LaunchWhenDone
)

function Test-IsOsArchX64 {
	if ($PSVersionTable.PSVersion.Major -lt 6) {
		return (Get-CimInstance -ClassName Win32_OperatingSystem).OSArchitecture -match '64.bit.?'
	}

	return [System.Runtime.InteropServices.RuntimeInformation]::OSArchitecture -eq [System.Runtime.InteropServices.Architecture]::X64
}

function DownloadFile($url, $targetFile) {
	$uri = New-Object "System.Uri" "$url"
	
	$request = [System.Net.HttpWebRequest]::Create($uri)

	$request.set_Timeout(15000) #15 second timeout

	$response = $request.GetResponse()

	$totalLength = [System.Math]::Floor($response.get_ContentLength() / 1024)

	$responseStream = $response.GetResponseStream()

	$targetStream = New-Object -TypeName System.IO.FileStream -ArgumentList $targetFile, Create

	$buffer = new-object byte[] 10KB

	$count = $responseStream.Read($buffer, 0, $buffer.length)

	$downloadedBytes = $count

	while ($count -gt 0) {
		$targetStream.Write($buffer, 0, $count)

		$count = $responseStream.Read($buffer, 0, $buffer.length)

		$downloadedBytes = $downloadedBytes + $count

		Write-Progress -activity "Downloading file '$($url.split('/') | Select-Object -Last 1)'" -status "Downloaded ($([System.Math]::Floor($downloadedBytes/1024))K of $($totalLength)K): " -PercentComplete ((([System.Math]::Floor($downloadedBytes / 1024)) / $totalLength) * 100)
	}

	Write-Progress -activity "Finished downloading file '$($url.split('/') | Select-Object -Last 1)'"  -Status "Ready" -Completed

	$targetStream.Flush()

	$targetStream.Close()

	$targetStream.Dispose()

	$responseStream.Dispose()
}

# Check running correct script for OS
if (($IsLinux -or $IsMacOS)) {
	throw "Running Sage X3 Automated Test Platform Client install for Windows on MacOS! Run ATPClientInstallMacOS.ps1 instead"
}

if (-NOT ([Security.Principal.WindowsPrincipal] [Security.Principal.WindowsIdentity]::GetCurrent()).IsInRole([Security.Principal.WindowsBuiltInRole] "Administrator")) {
	write-Warning "This setup needs admin permissions. Please run this file as admin."     
	break
}

#Legal notice
Write-host "`nThe Sage X3 Automated Test Platform Installer facilitates your download of the below listed OS components from source as well as Microsoft Visual Studio 2019 software.
The license information for the OS components are as follows:`n
Git - https://git-scm.com/about/free-and-open-source; `n
Node.js - https://raw.githubusercontent.com/nodejs/node/master/LICENSE;`n
and Microsoft Visual Studio Code - https://code.visualstudio.com/license. `n
For the license information of Microsoft Visual Studio 2019 software visit  https://visualstudio.microsoft.com/license-terms/mlt031519/;`n
You are responsible for compliance with these license terms. The Sage X3 Automated Test Platform Installer will allow you to download a binary version of the open source code. 
If you require the source code then this can be downloaded separately from the following link: `n
Git- https://github.com/git/git.`n
Please find the license information for Automated Test Platform Tool (`"ATPT`"):
The Sage X3 Automated Test Platform Tool (`"ATPT`") comprises two types of components: The Sage Proprietary Code (`"SPC`") and components that are supplied and licensed by third parties (`"The Third-Party Services`"), which are listed in the documentation ATP-Third-Party-Components. From time to time and at Sage's discretion, additional SPC may be provided for use alongside ATPT.
The SPC is licensed to you in accordance with the terms of the Sage X3 End User License Agreement (EULA). Your use of the Third-Party Services is subject to the license terms and conditions imposed by the relevant third parties. If you do not accept or agree to these third-party terms and conditions, you might not be able to operate the ATPT, or the ATPT Installer. You are responsible for reviewing the applicable third-party terms and conditions including all applicable data privacy and data gathering policies and you should make whatever investigation you feel necessary or appropriate before proceeding with any transaction related to Third-Party Services. You are responsible for your compliance with the license terms.
Sage does not endorse, and does not make any representation, warranty or promise regarding the Third-Party Services and shall have no liability whatsoever for any damage, liabilities or losses caused by any Third Party Services, regardless of whether the Third-Party Services are described as `"authorised`", `"certified`", `"recommended`" or the like.
Sage has not developed any of the Third Party Services and has no obligation to provide any support for Third Party Services and does not guarantee the initial or continuing interoperability of the ATPT with the Third-Party Services. If any third-party ceases to make Third-Party Services available for interoperation with the ATPT, Sage may cease providing the ATPT or may choose to operate with alternative Third-Party Services without any further responsibility or liability to you.`n"

$confirmation = read-host "Do you wish to continue? [y/N]"
if ($confirmation -ne "y") {
	return
}

# Which applications are already installed
$installNode = $TRUE
$installGit = $FALSE
$installVSCode = $TRUE

#In case user installs Java or Chrome after running the script once without reloading Powershell
$env:Path = [System.Environment]::GetEnvironmentVariable("Path", "Machine") + ";" + [System.Environment]::GetEnvironmentVariable("Path", "User")

if (Get-Command java -errorAction SilentlyContinue) {
	$current_version = (java -version)
	write-host "Java version $current_version installed"
}
else {
	write-host "Java not installed. Stopping install." -ForegroundColor Red
	break
}

try {
	$chromeVersion = (Get-Item (Get-ItemProperty 'HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\chrome.exe' -errorAction SilentlyContinue).'(Default)' -errorAction SilentlyContinue).VersionInfo.ProductVersion
}
catch {
}
if ($chromeVersion) {
	write-host "Chrome version $chromeVersion installed"
}
else {
	write-host "Chrome not installed. Stopping install." -ForegroundColor Red
	break
}

switch ($Architecture) {
	'64-bit' {
		if (Test-IsOsArchX64) {
			break
		}

		Write-Warning '64-bit install requested on 32-bit system. Switching to 32-bit'
		$Architecture = '32-bit'
		break
	}
}

[Net.ServicePointManager]::SecurityProtocol = [Net.SecurityProtocolType]::Tls12

$confirmation = read-host "Do you want to install Git? [y/N]"
if ($confirmation -eq "y") {
	$installGit = $TRUE
	if (Get-Command git -errorAction SilentlyContinue) {
		$current_version = (git --version)
		write-host "Git $current_version already installed" -ForegroundColor Yellow
		$confirmation = read-host "Are you sure you want to replace this version ? [y/N]"
		if ($confirmation -ne "y") {
			$installGit = $FALSE
		}
	}
}

$tmpdir = [System.IO.Path]::GetTempPath()

if ($installGit) {
	try {
		write-host "Downloading Git"

		$gitresponse = Invoke-WebRequest -URI https://github.com/git-for-windows/git/releases/latest/ -UseBasicParsing
		$gitReleaseLink = $gitresponse.Links | Where-Object { $_.Href -like "/git-for-windows/git/releases/tag/v*" } | Select-Object -Expand Href | Out-String
		$gitReleaseVersion = $gitReleaseLink.Substring($gitReleaseLink.LastIndexOf("/") + 1).Trim()
		if ($gitReleaseVersion -match 'windows') {
			$gitVersionNo = $gitReleaseVersion.Substring(1, $gitReleaseVersion.IndexOf(".win") - 1)
			$gitSubVersion = $gitReleaseVersion.Substring($gitReleaseVersion.IndexOf(".win"))
			if ($gitSubVersion.Substring($gitSubVersion.LastIndexOf(".") + 1) -gt 1) {
				$gitVersionNo = $gitVersionNo + "." + $gitSubVersion.Substring($gitSubVersion.LastIndexOf(".") + 1)
			}
		}
		else {
			$gitVersionNo = $gitReleaseVersion.Substring(1)
		}
		$gitURL = "https://github.com/git-for-windows/git/releases/download/$gitReleaseVersion/Git-$gitVersionNo-$Architecture.exe"
		$index = $gitURL.LastIndexOf("/")
		$installerName = $gitURL.Substring($index + 1).Trim()
		$installerPath = [System.IO.Path]::Combine($tmpdir, $installerName)

		DownloadFile $gitURL $installerPath

		write-Output "$installerName downloaded"
        
		write-host "Installing Git from $installerPath" -ForegroundColor Yellow
		Start-Process "$installerPath" -ArgumentList '/VERYSILENT /NORESTART /NOCANCEL /SP- /CLOSEAPPLICATIONS /RESTARTAPPLICATIONS /COMPONENTS="icons,ext\reg\shellhere,assoc,assoc_sh"' -Wait -NoNewWindow
		
		$env:Path = [System.Environment]::GetEnvironmentVariable("Path", "Machine") + ";" + [System.Environment]::GetEnvironmentVariable("Path", "User") 
		
		$gitInstalledVersion = (git --version)
		
		write-host "Git version $gitInstalledVersion Installed successfully.`n`n" -ForegroundColor Green
		
		Remove-Item $installerPath
	}
	catch {
		write-host $_ -ForegroundColor Red
		write-host "Git Install failed`n`n" -ForegroundColor Red
	}
}

Write-Host "Checking Nodejs versions" -ForegroundColor Yellow

$nodeDownloadLink = "https://nodejs.org/dist/"
$baseNodeVersion = "v12"

$response = Invoke-WebRequest -URI $nodeDownloadLink -UseBasicParsing

$nodeversion = $response.Links | Where-Object { $_.Href -like "$baseNodeVersion.*.*" } | Select-Object -Expand Href |  Sort-Object -Property @{Expression = { [int]$_.replace('/', '').split('.')[1] }; Descending = $True }, @{Expression = { [int]$_.replace('/', '').split('.')[2] }; Descending = $True } | Select-Object -First 1 | Out-String

$nodeversion = $nodeversion.replace("`n", "").replace("`r", "").Replace("/", "")

if (Get-Command node -errorAction SilentlyContinue) {
	$current_version = (node -v)
	write-host "Nodejs $current_version already installed" -ForegroundColor Yellow

	$installedVersionNumbers = $current_version.replace("v", "").split(".")
	$onlineVersionNumbers = $nodeversion.Replace("v", "").Split(".")

	$confirmation = "n"
	if ([int]$installedVersionNumbers[0] -gt [int]$onlineVersionNumbers[0]) {
		Write-Host "Sage X3 Automated Test Platform is not compatible with Nodejs version $current_version and requires Nodejs $baseNodeVersion.
Please uninstall Nodejs $current_version and run the installer again.
Before uninstalling Nodejs, please note, that changing from Nodejs version $current_version may break compatibility with other Nodejs applications if they require that specific version." -ForegroundColor Red -BackgroundColor Black
		return
	}
	elseif ([int]$installedVersionNumbers[0] -lt [int]$onlineVersionNumbers[0]) {
		Write-Warning "`nSage X3 Automated Test Platform is not compatible with Nodejs version $current_version and requires Nodejs $baseNodeVersion.
Please note, that upgrading from Nodejs version $current_version may break compatibility with other Nodejs applications if they require that specific version.
Please note, If you decide not to update Nodejs, the installation will stop."
		$confirmation = read-host "Are you sure you want to replace Nodejs $current_version with $nodeversion ? [y/N]"
		if ($confirmation -ne "y") {
			return
		}
	}
	else {
		if ([int]$installedVersionNumbers[1] -le [int]$onlineVersionNumbers[1]) {
			if ([int]$installedVersionNumbers[1] -lt [int]$onlineVersionNumbers[1] -or [int]$installedVersionNumbers[2] -lt [int]$onlineVersionNumbers[2]) {
				Write-Host "Newer version of Nodejs detected" -ForegroundColor Yellow
				$confirmation = read-host "Would you like to upgrade to Nodejs $nodeversion ? [y/N]"
				if ($confirmation -ne "y") {
					$installNode = $FALSE
				}
			}
			else {
				Write-Host "Nodejs already up to date" -ForegroundColor Green
				$installNode = $FALSE
			}
		}
	}
}

if ($installNode) {
	write-host "Downloading NodeJs"

	if ($Architecture -ne '32-bit') {
		$Bits = "x64"
	}
	else {
		$Bits = "x86"
	}
	
	$response = Invoke-WebRequest -URI $nodeDownloadLink$nodeversion -UseBasicParsing

	$nodeInstaller = $response.Links | Where-Object { $_.Href -like "*node-v*-$Bits.msi" } | Select-Object -Expand Href | Out-String 
	
	$nodeUrl = "$nodeDownloadLink$nodeversion/$nodeInstaller"
		
	try {
		$installerName = "node-install.msi"

		$installerPath = [System.IO.Path]::Combine($tmpdir, $installerName)

		DownloadFile $nodeUrl $installerPath

		write-Output "$installerName downloaded"

		Write-host "Installing Nodejs from $installerPath" -ForegroundColor Yellow
		Start-Process "msiexec.exe" -ArgumentList "/i $installerPath /passive" -Wait -NoNewWindow
		
		$env:Path = [System.Environment]::GetEnvironmentVariable("Path", "Machine") + ";" + [System.Environment]::GetEnvironmentVariable("Path", "User") 
		
		$nodeInstalledVersion = (node -v)
		
		write-host "Node version $nodeInstalledVersion Installed successfully.`n`n" -ForegroundColor Green
		
		Remove-Item $installerPath
	}
	catch {
		write-host $_ -ForegroundColor Red
		throw "Node installation Failed!"
	}
}

Write-Host "Checking Visual Studio Code versions" -ForegroundColor Yellow

if (Get-Command code -errorAction SilentlyContinue) {
	$current_version = (code -v).split("\n") | Select-Object -First 1
	Write-Host "VS Code $current_version already installed" -ForegroundColor Yellow

	$confirmation = Read-Host "Would you like to upgrade to VS Code $vsCodeVersion ? [y/N]"
	if ($confirmation -ne "y") {
		$installVSCode = $FALSE
	}
}

if ($installVSCode) {
	try {
		Write-Host "Downloading VS Code..." -ForegroundColor Yellow
		$prevProgressPreference = $ProgressPreference
		$ProgressPreference = 'SilentlyContinue'

		# Get information required for installation

		if ($Architecture -ne '32-bit') {
			$platform = "win32-x64"
		}
		else {
			$platform = "win32"
		}
		
		$installerName = "vscode-install.exe"

		$installerPath = [System.IO.Path]::Combine($tmpdir, $installerName)
		
		$FileUri = "https://aka.ms/$platform-user-stable"
		
		DownloadFile $FileUri $installerPath

		Start-Sleep -s 5 

		$exeArgs = '/verysilent /tasks=addtopath'
		if ($EnableContextMenus) {
			$exeArgs = '/verysilent /tasks=addcontextmenufiles,addcontextmenufolders,addtopath'
		}

		if (-not $PSCmdlet.ShouldProcess("$installerPath $exeArgs", 'Start-Process -Wait')) {
			break
		}

		Write-Host "Installing VS Code...`n" -ForegroundColor Yellow

		Start-Process -Wait $installerPath -ArgumentList $exeArgs

		$codeExePath = "$env:LOCALAPPDATA\Programs\Microsoft VS Code\bin\code.cmd"

		# Install any extensions
		$extensions = @("alexkrechik.cucumberautocomplete", "rebornix.project-snippets", "streetsidesoftware.code-spell-checker", "eg2.tslint", "esbenp.prettier-vscode", "vscode-icons-team.vscode-icons")
		if ($PSCmdlet.ShouldProcess(($extensions -join ','), "$codeExePath --install-extension")) {
			foreach ($extension in $extensions) {
				Write-Host "`nInstalling extension $extension..." -ForegroundColor Yellow
				& $codeExePath --install-extension $extension
			}
		}
		$env:Path = [System.Environment]::GetEnvironmentVariable("Path", "Machine") + ";" + [System.Environment]::GetEnvironmentVariable("Path", "User")

		if ($PSCmdlet.ShouldProcess('VS Code installation complete!', 'Write-Host')) {
			$currentCodeVersion = (code -v)
			Write-Host "VS Code version $currentCodeVersion installed`n" -ForegroundColor Green
		}
	}
	catch {
		write-host $_ -ForegroundColor Red
		throw "VS Code installation Failed!"
	}
	finally {
		$ProgressPreference = $prevProgressPreference
	}
}

Write-Host "Download and install Automated Test Platform`n"

$folderMessage = "Enter folder to install Sage X3 Automated Test Platform to (leave blank to use current folder)"

if ($SkipFolderInstall) {
	$folderMessage = "Enter folder Sage X3 Automated Test Platform is installed in (leave blank to use current folder)"
}

$atpFolder = read-host $folderMessage

if ($atpFolder -eq "") {
	$atpFolder = "$PSScriptRoot"
}

if (-not $SkipFolderInstall) {
	if (-NOT (Test-Path -Path $atpFolder)) {
		New-Item -ItemType Directory -Path "$atpFolder" -Force > $null
	}
}

Set-Location $atpFolder

if (-not $SkipATPInstall) {
	try {
		if ((Test-Path "${env:ProgramFiles(x86)}\\Microsoft Visual Studio\\2019\\BuildTools") -or (Test-Path "${env:ProgramFiles}\\Microsoft Visual Studio\\2019\\BuildTools")) {
			Write-host "Microsoft Visual Studio Build tools installed" -ForegroundColor Green
		}
		else {
			#install vs build tools
			$installerName = "vs-buildtools.exe"

			$installerPath = [System.IO.Path]::Combine($tmpdir, $installerName)

			Write-host "Downloading VS Build Tools..." -ForegroundColor Yellow
			DownloadFile "https://aka.ms/vs/16/release/vs_buildtools.exe" "$installerPath"
			Write-host "$installerName downloaded"

			Write-Host "Installing VS Build Tools...
			Please wait. This may take some time`n" -ForegroundColor Yellow
			$exeArgs = "--quiet --norestart --nocache"
			Start-Process -Wait $installerPath -ArgumentList $exeArgs
			npm config set msvs_version 2017
			Write-Host "`nVS Build Tools installation complete!`n`n" -ForegroundColor Green
		}
	}
	catch {
		Write-Host $_ -ForegroundColor Red
		throw "Error installing VS Build Tools"
	}

	try {
		if (-not (Get-Command npm-cli-login -errorAction SilentlyContinue)) {
			npm install -g npm-cli-login
		}
		Start-Process -Wait "npm" -ArgumentList "config set @sageatp:registry https://repository.sagex3.com:8443/repository/atp-npm/"
		npm config set always-auth true
		npm config set strict-ssl false

		$loginAttempt = 0
		$loginSuccess = $FALSE
		Do {
			$loginAttempt++
			if ($loginAttempt -gt 3) {
				Write-Error "Number of login attempts exceeded. Cancelling installation."
				return
			}
			$username = Read-host "Enter Sage X3 Automated Test Platform Username"
			$password = Read-Host "Enter Sage X3 Automated Test Platform Password" -AsSecureString

			#npm-cli-login only accepts plain strings as input. 
			$notSecure = [System.Runtime.InteropServices.Marshal]::PtrToStringAuto([System.Runtime.InteropServices.Marshal]::SecureStringToBSTR($password))
			$loginResult = npm-cli-login -u "$username" -p "$notSecure" -e "$username@sage.com" -r https://repository.sagex3.com:8443/repository/atp-npm -s '@sageatp' 2>&1
			$notSecure = $NULL

			#npm-cli-login does not cause an error when it fails. This checks the output for any errors.
			write-host $loginResult
			$checkStrings = @('Incorrect username or password', 'Bad username or password', 'throw new Error')
			if ($null -ne ($checkStrings | Where-Object { $loginResult -match $_ })) {
				Write-Host "Login to Sage X3 Automated Test Platform repository failed. Please check your user credentials and try again.`n" -ForegroundColor Red
			}
			else {
				Write-Host "Login to Sage X3 Automated Test Platform repository succeeded" -ForegroundColor Green
				$loginSuccess = $TRUE
			}
		} while (-not $loginSuccess)

		$newVersion = "v2.7.0"

		Write-host "Installing Sage X3 Automated Test Platform $newVersion ..." -ForegroundColor Yellow
		npm install @sageatp/x3-atppackage@2.7.0 --ignore-scripts

		XCopy "$atpfolder\node_modules\@sageatp\x3-atppackage" "$atpFolder\" /s /y /q > $null

		Remove-Item -Path "$atpfolder\node_modules\@sageatp\x3-atppackage" -Recurse

		npm install

		Copy-Item -Path "$atpfolder\cucumber-config\parameters-cloud-chrome_template" -Destination "$atpfolder\cucumber-config\parameters-cloud-chrome"

		Write-Host "Sage X3 Automated Test Platform $newVersion installed successfully!`n" -ForegroundColor Green
	}
	catch {
		Write-Host $_ -ForegroundColor Red
		throw "Error installing npm packages"
	}
}

# Launch if requested
if ($LaunchWhenDone) {
	Write-Host "`nInstallation complete, starting VS Code...`n`n" -ForegroundColor Green
	& code .
	return
}

Write-Host "`nInstallation complete!`n`n" -ForegroundColor Green